from ansible import utils
from ansible.inventory import Group
from ansible.inventory import Host as BaseHost
from ansible.inventory import Inventory as BaseInventory
import inspect
import logging


log = logging.getLogger('ploy_ansible.inventory')


class PloyInventoryDict(dict):
    def __getitem__(self, name):
        if name.startswith('awsome_'):
            caller_frame = inspect.currentframe().f_back
            info = inspect.getframeinfo(caller_frame)
            new_name = "ploy_%s" % name[7:]
            log.warning("Use of deprecated variable name '%s', use '%s' instead.\n%s:%s\n%s" % (
                name, new_name, info.filename, info.lineno, ''.join(info.code_context)))
        return dict.__getitem__(self, name)


class Host(BaseHost):
    def __init__(self, ctrl, name):
        BaseHost.__init__(self, name)
        self.ctrl = ctrl

    @property
    def vars(self):
        instance = self.ctrl.instances[self.name]
        results = dict(
            self._ploy_vars,
            ansible_connection='execnet_connection',
            ansible_ssh_user=instance.config.get('user', 'root'),
            _ploy_instance=instance,
            _ploy_instances=self.ctrl.instances)
        for k, v in instance.config.items():
            if k == 'password' and instance.config['password-fallback']:
                results['ansible_ssh_pass'] = v
            elif k.startswith('ansible_'):
                results[k] = v
            elif k.startswith('ansible-'):
                results[k[len('ansible-'):].replace('-', '_')] = v
            else:
                results['ploy_%s' % k.replace('-', '_')] = v
                results['awsome_%s' % k.replace('-', '_')] = v
        return results

    @vars.setter
    def vars(self, value):
        self._ploy_vars = value


class Inventory(BaseInventory):
    def __init__(self, ctrl, vault_password=None):
        from ploy_ansible import get_playbooks_directory
        kwargs = dict(host_list=[])
        if vault_password is not None:
            kwargs['vault_password'] = vault_password
        BaseInventory.__init__(self, **kwargs)
        self.ctrl = ctrl
        groups = {}
        groups['all'] = self.get_group('all')
        seen = set()
        for instance in self.ctrl.instances.values():
            if instance.uid in seen:
                continue
            seen.add(instance.uid)
            h = Host(ctrl, instance.uid)
            add_to = ['all', '%ss' % instance.sectiongroupname]
            if hasattr(instance, 'master'):
                master = instance.master
                if instance == getattr(master, 'instance', None):
                    add_to.append('masters')
                else:
                    add_to.append('%s-instances' % master.id)
            add_to.extend(instance.config.get('groups', '').split())
            for group in add_to:
                g = groups.get(group)
                if g is None:
                    g = self.get_group(group)
                    if g is None:
                        g = Group(group)
                        self.add_group(g)
                    groups[group] = g
                if h not in g.hosts:
                    g.add_host(h)
        self._vars_plugins = [x for x in utils.plugins.vars_loader.all(self)]
        self._hosts_cache.clear()
        self._pattern_cache.clear()
        self.set_playbook_basedir(get_playbooks_directory(ctrl.config))

    def get_variables(self, hostname, **kwargs):
        result = BaseInventory.get_variables(self, hostname, **kwargs)
        if hasattr(self, 'get_host_variables'):
            result = utils.combine_vars(
                result,
                self.get_host_variables(
                    hostname, vault_password=self._vault_password))
        return PloyInventoryDict(result)



from ansible.module_utils._text import to_text
from ansible.plugins.lookup import LookupBase


class LookupModule(LookupBase):
    def run(self, terms, variables, **kwargs):

        ret = []

        for term in terms:
            lookupfile = self.find_file_in_search_path(variables, 'files', term)
            if lookupfile:
                (b_contents, show_data) = self._loader._get_file_contents(lookupfile)
                vaultlib = variables['_ploy_instance'].get_vault_lib()
                if vaultlib.is_encrypted(b_contents):
                    b_contents = vaultlib.decrypt(b_contents)
                b_contents = to_text(b_contents, errors='surrogate_or_strict')
                ret.append(b_contents)

        return ret



from __future__ import print_function
import os
import sys


def connect_patch_factory(ctrl):
    def connect_patch(self, *args, **kwargs):
        self.runner._ploy_ctrl = ctrl
        return self._ploy_orig_connect(*args, **kwargs)
    return connect_patch


def patch_connect(ctrl):
    from ansible.runner.connection import Connector
    if not hasattr(Connector, '_ploy_orig_connect'):
        Connector._ploy_orig_connect = Connector.connect
        Connector.connect = connect_patch_factory(ctrl)


def parse_extra_vars(extras, vault_pass=None):
    from ansible import utils
    extra_vars = {}
    for extra_vars_opt in extras:
        if extra_vars_opt.startswith("@"):
            # Argument is a YAML file (JSON is a subset of YAML)
            kw = {}
            if vault_pass:
                kw['vault_password'] = vault_pass
            extra_vars = utils.combine_vars(extra_vars, utils.parse_yaml_from_file(extra_vars_opt[1:]), **kw)
        elif extra_vars_opt and extra_vars_opt[0] in '[{':
            # Arguments as YAML
            extra_vars = utils.combine_vars(extra_vars, utils.parse_yaml(extra_vars_opt))
        else:
            # Arguments as Key-value
            extra_vars = utils.combine_vars(extra_vars, utils.parse_kv(extra_vars_opt))
    return extra_vars


def get_vault_password_source(main_config, option='vault-password-source'):
    from ploy_ansible import KeyringSource, NullSource
    ansible_config = main_config.get('global', {}).get('ansible', {})
    src = ansible_config.get(option)
    if src is None:
        return NullSource()
    return KeyringSource(src)


def ansible_cmd(ctrl, argv):
    import ansible.constants as C
    from ansible.runner import Runner
    from ansible import errors
    from ansible import callbacks
    from ploy_ansible.inventory1 import Inventory
    from ansible import utils
    from ansible.utils.vault import VaultLib

    parser = utils.base_parser(
        constants=C,
        runas_opts=True,
        subset_opts=True,
        output_opts=True,
        check_opts=True,
        diff_opts=False,
        usage='%s ansible <host-pattern> [options]' % ctrl.progname
    )
    parser.remove_option('-i')
    parser.remove_option('-k')
    parser.add_option(
        '-a', '--args', dest='module_args',
        help="module arguments", default=C.DEFAULT_MODULE_ARGS)
    parser.add_option(
        '-m', '--module-name', dest='module_name',
        help="module name to execute (default=%s)" % C.DEFAULT_MODULE_NAME,
        default=C.DEFAULT_MODULE_NAME)
    options, args = parser.parse_args(argv)
    if len(args) == 0 or len(args) > 1:
        parser.print_help()
        sys.exit(1)

    if hasattr(options, 'become_ask_pass'):
        # privlege escalation command line arguments need to be mutually exclusive
        utils.check_mutually_exclusive_privilege(options, parser)
    else:
        # su and sudo command line arguments need to be mutually exclusive
        if (hasattr(options, 'su')
                and (options.su or options.su_user or options.ask_su_pass)
                and (options.sudo or options.sudo_user or options.ask_sudo_pass)):
            parser.error("Sudo arguments ('--sudo', '--sudo-user', and '--ask-sudo-pass') "
                         "and su arguments ('-su', '--su-user', and '--ask-su-pass') are "
                         "mutually exclusive")

    if hasattr(options, 'ask_vault_pass') and (options.ask_vault_pass and options.vault_password_file):
        parser.error("--ask-vault-pass and --vault-password-file are mutually exclusive")

    cbs = callbacks.CliRunnerCallbacks()
    cbs.options = options
    pattern = args[0]
    patch_connect(ctrl)
    vault_pass = None
    kw = {}
    if hasattr(options, 'become_ask_pass'):
        becomepass = None
        become_method = None
        utils.normalize_become_options(options)
        become_method = utils.choose_pass_prompt(options)
        kw['become_ask_pass'] = options.become_ask_pass
        kw['become_method'] = become_method
    else:
        sudopass = None
        su_pass = None
        options.ask_sudo_pass = options.ask_sudo_pass or C.DEFAULT_ASK_SUDO_PASS
        kw['ask_sudo_pass'] = options.ask_sudo_pass
        if hasattr(options, 'ask_su_pass'):
            options.ask_su_pass = options.ask_su_pass or C.DEFAULT_ASK_SU_PASS
            kw['ask_su_pass'] = options.ask_sudo_pass
    if hasattr(options, 'ask_vault_pass'):
        options.ask_vault_pass = options.ask_vault_pass or C.DEFAULT_ASK_VAULT_PASS
        kw['ask_vault_pass'] = options.ask_vault_pass
    passwds = utils.ask_passwords(**kw)
    if hasattr(options, 'become_ask_pass'):
        (sshpass, becomepass, vault_pass) = passwds
    else:
        if len(passwds) == 2:
            (sshpass, sudopass) = passwds
        elif len(passwds) == 3:
            (sshpass, sudopass, su_pass) = passwds
        else:
            (sshpass, sudopass, su_pass, vault_pass) = passwds
    if VaultLib is not None and vault_pass is None:
        vault_pass = get_vault_password_source(ctrl.config).get()
    if getattr(options, 'vault_password_file', None):
        this_path = os.path.expanduser(options.vault_password_file)
        try:
            f = open(this_path, "rb")
            tmp_vault_pass = f.read().strip()
            f.close()
        except (OSError, IOError) as e:
            raise errors.AnsibleError("Could not read %s: %s" % (this_path, e))

        if not options.ask_vault_pass:
            vault_pass = tmp_vault_pass

    inventory_manager = Inventory(ctrl, vault_password=vault_pass)
    if options.subset:
        inventory_manager.subset(options.subset)
    hosts = inventory_manager.list_hosts(pattern)
    if len(hosts) == 0:
        callbacks.display("No hosts matched", stderr=True)
        sys.exit(1)
    if options.listhosts:
        for host in hosts:
            callbacks.display('    %s' % host)
        sys.exit(0)
    if options.module_name in ['command', 'shell'] and not options.module_args:
        callbacks.display("No argument passed to %s module" % options.module_name, color='red', stderr=True)
        sys.exit(1)

    if not hasattr(options, 'become_ask_pass'):
        if options.sudo_user or options.ask_sudo_pass:
            options.sudo = True
        options.sudo_user = options.sudo_user or C.DEFAULT_SUDO_USER
        if hasattr(options, 'su'):
            if options.su_user or options.ask_su_pass:
                options.su = True
            options.su_user = options.su_user or C.DEFAULT_SU_USER
    if options.tree:
        utils.prepare_writeable_dir(options.tree)
    kw = {}
    if hasattr(options, 'become_ask_pass'):
        kw['become'] = options.become
        kw['become_method'] = options.become_method
        kw['become_pass'] = becomepass
        kw['become_user'] = options.become_user
    else:
        if hasattr(options, 'su'):
            kw['su'] = options.su
            kw['su_user'] = options.su_user
        if hasattr(options, 'su_pass'):
            kw['su_pass'] = options.su_pass
        kw['sudo'] = options.sudo
        kw['sudo_user'] = options.sudo_user
        kw['sudo_pass'] = sudopass
    if vault_pass:
        kw['vault_pass'] = vault_pass
    runner = Runner(
        module_name=options.module_name,
        module_path=options.module_path,
        module_args=options.module_args,
        remote_user=options.remote_user,
        inventory=inventory_manager,
        timeout=options.timeout,
        private_key_file=options.private_key_file,
        forks=options.forks,
        pattern=pattern,
        callbacks=cbs,
        transport='ssh',
        subset=options.subset,
        check=options.check,
        diff=options.check,
        **kw)
    results = runner.run()
    for result in results['contacted'].values():
        if 'failed' in result or result.get('rc', 0) != 0:
            sys.exit(2)
    if results['dark']:
        sys.exit(3)


def playbook_cmd(ctrl, argv):
    import ansible.playbook
    import ansible.constants as C
    from ansible import errors
    from ansible import callbacks
    from ansible import utils
    from ansible.color import ANSIBLE_COLOR, stringc
    from ansible.utils.vault import VaultLib
    from ploy_ansible import ansible_version
    from ploy_ansible.inventory1 import Inventory

    parser = utils.base_parser(
        constants=C,
        connect_opts=True,
        runas_opts=True,
        subset_opts=True,
        check_opts=True,
        diff_opts=True,
        usage='%s playbook playbook.yml' % ctrl.progname
    )
    parser.remove_option('-i')
    parser.remove_option('-k')
    if not parser.has_option('--extra-vars'):
        parser.add_option(
            '-e', '--extra-vars', dest="extra_vars", action="append",
            help="set additional variables as key=value or YAML/JSON", default=[])
    parser.add_option(
        '-t', '--tags', dest='tags', default='all',
        help="only run plays and tasks tagged with these values")
    parser.add_option(
        '--skip-tags', dest='skip_tags',
        help="only run plays and tasks whose tags do not match these values")
    parser.add_option(
        '--syntax-check', dest='syntax', action='store_true',
        help="perform a syntax check on the playbook, but do not execute it")
    parser.add_option(
        '--list-tasks', dest='listtasks', action='store_true',
        help="list all tasks that would be executed")
    parser.add_option(
        '--step', dest='step', action='store_true',
        help="one-step-at-a-time: confirm each task before running")
    parser.add_option(
        '--start-at-task', dest='start_at',
        help="start the playbook at the task matching this name")
    if ansible_version >= (1, 6):
        parser.add_option(
            '--force-handlers', dest='force_handlers', action='store_true',
            help="run handlers even if a task fails")
    options, args = parser.parse_args(argv)
    cbs = callbacks.CliRunnerCallbacks()
    cbs.options = options
    if len(args) == 0:
        parser.print_help(file=sys.stderr)
        sys.exit(1)

    if hasattr(options, 'become_ask_pass'):
        # privlege escalation command line arguments need to be mutually exclusive
        utils.check_mutually_exclusive_privilege(options, parser)
    else:
        # su and sudo command line arguments need to be mutually exclusive
        if (hasattr(options, 'su')
                and (options.su or options.su_user or options.ask_su_pass)
                and (options.sudo or options.sudo_user or options.ask_sudo_pass)):
            parser.error("Sudo arguments ('--sudo', '--sudo-user', and '--ask-sudo-pass') "
                         "and su arguments ('-su', '--su-user', and '--ask-su-pass') are "
                         "mutually exclusive")

    if hasattr(options, 'ask_vault_pass') and (options.ask_vault_pass and options.vault_password_file):
        parser.error("--ask-vault-pass and --vault-password-file are mutually exclusive")

    def colorize(lead, num, color):
        """ Print 'lead' = 'num' in 'color' """
        if num != 0 and ANSIBLE_COLOR and color is not None:
            return "%s%s%-15s" % (stringc(lead, color), stringc("=", color), stringc(str(num), color))
        else:
            return "%s=%-4s" % (lead, str(num))

    def hostcolor(host, stats, color=True):
        if ANSIBLE_COLOR and color:
            if stats['failures'] != 0 or stats['unreachable'] != 0:
                return "%-37s" % stringc(host, 'red')
            elif stats['changed'] != 0:
                return "%-37s" % stringc(host, 'yellow')
            else:
                return "%-37s" % stringc(host, 'green')
        return "%-26s" % host

    try:
        patch_connect(ctrl)
        if hasattr(options, 'become_ask_pass'):
            becomepass = None
        else:
            sudopass = None
            su_pass = None
        vault_pass = None
        if not options.listhosts and not options.syntax and not options.listtasks:
            kw = {}
            if hasattr(options, 'become_ask_pass'):
                utils.normalize_become_options(options)
                become_method = utils.choose_pass_prompt(options)
                kw['become_ask_pass'] = options.become_ask_pass
                kw['become_method'] = become_method
            else:
                options.ask_sudo_pass = options.ask_sudo_pass or C.DEFAULT_ASK_SUDO_PASS
                kw['ask_sudo_pass'] = options.ask_sudo_pass
                if hasattr(options, 'ask_su_pass'):
                    options.ask_su_pass = options.ask_su_pass or C.DEFAULT_ASK_SU_PASS
                    kw['ask_su_pass'] = options.ask_sudo_pass
            if hasattr(options, 'ask_vault_pass'):
                options.ask_vault_pass = options.ask_vault_pass or C.DEFAULT_ASK_VAULT_PASS
                kw['ask_vault_pass'] = options.ask_vault_pass
            passwds = utils.ask_passwords(**kw)
            if hasattr(options, 'become_ask_pass'):
                (sshpass, becomepass, vault_pass) = passwds
            else:
                if len(passwds) == 2:
                    (sshpass, sudopass) = passwds
                elif len(passwds) == 3:
                    (sshpass, sudopass, su_pass) = passwds
                else:
                    (sshpass, sudopass, su_pass, vault_pass) = passwds
            if VaultLib is not None and vault_pass is None:
                vault_pass = get_vault_password_source(ctrl.config).get()
            if options.sudo_user or options.ask_sudo_pass:
                options.sudo = True
            options.sudo_user = options.sudo_user or C.DEFAULT_SUDO_USER
            if hasattr(options, 'su'):
                if options.su_user or options.ask_su_pass:
                    options.su = True
                options.su_user = options.su_user or C.DEFAULT_SU_USER
            if getattr(options, 'vault_password_file', None):
                this_path = os.path.expanduser(options.vault_password_file)
                try:
                    f = open(this_path, "rb")
                    tmp_vault_pass = f.read().strip()
                    f.close()
                except (OSError, IOError) as e:
                    raise errors.AnsibleError("Could not read %s: %s" % (this_path, e))

                if not options.ask_vault_pass:
                    vault_pass = tmp_vault_pass
        inventory = Inventory(ctrl, vault_password=vault_pass)
        extra_vars = parse_extra_vars(options.extra_vars, vault_pass=vault_pass)
        only_tags = options.tags.split(",")
        skip_tags = options.skip_tags
        if options.skip_tags is not None:
            skip_tags = options.skip_tags.split(",")

        for playbook in args:
            if not os.path.exists(playbook):
                raise errors.AnsibleError("the playbook: %s could not be found" % playbook)

        # run all playbooks specified on the command line
        for playbook in args:
            playbook = os.path.abspath(playbook)

            # let inventory know which playbooks are using so it can know the basedirs
            inventory.set_playbook_basedir(os.path.dirname(playbook))

            stats = callbacks.AggregateStats()
            playbook_cb = callbacks.PlaybookCallbacks(verbose=utils.VERBOSITY)
            if options.step:
                playbook_cb.step = options.step
            if options.start_at:
                playbook_cb.start_at = options.start_at
            runner_cb = callbacks.PlaybookRunnerCallbacks(stats, verbose=utils.VERBOSITY)

            kw = {}
            if hasattr(options, 'become_ask_pass'):
                kw['become'] = options.become
                kw['become_method'] = options.become_method
                kw['become_pass'] = becomepass
                kw['become_user'] = options.become_user
            else:
                if hasattr(options, 'su'):
                    kw['su'] = options.su
                    kw['su_user'] = options.su_user
                if hasattr(options, 'su_pass'):
                    kw['su_pass'] = options.su_pass
                kw['sudo'] = options.sudo
                kw['sudo_user'] = options.sudo_user
                kw['sudo_pass'] = sudopass
            if vault_pass:
                kw['vault_password'] = vault_pass
            if hasattr(options, 'force_handlers'):
                kw['force_handlers'] = options.force_handlers
            pb = ansible.playbook.PlayBook(
                playbook=playbook,
                module_path=options.module_path,
                inventory=inventory,
                forks=options.forks,
                remote_user=options.remote_user,
                callbacks=playbook_cb,
                runner_callbacks=runner_cb,
                stats=stats,
                timeout=options.timeout,
                transport=options.connection,
                extra_vars=extra_vars,
                private_key_file=options.private_key_file,
                only_tags=only_tags,
                skip_tags=skip_tags,
                check=options.check,
                diff=options.diff,
                **kw)

            if options.listhosts or options.listtasks or options.syntax:
                print('')
                print('playbook: %s' % playbook)
                print('')
                playnum = 0
                for (play_ds, play_basedir) in zip(pb.playbook, pb.play_basedirs):
                    playnum += 1
                    play = ansible.playbook.Play(pb, play_ds, play_basedir)
                    label = play.name
                    hosts = pb.inventory.list_hosts(play.hosts)
                    # Filter all tasks by given tags
                    if pb.only_tags != 'all':
                        if options.subset and not hosts:
                            continue
                        matched_tags, unmatched_tags = play.compare_tags(pb.only_tags)

                        # Remove skipped tasks
                        matched_tags = matched_tags - set(pb.skip_tags)

                        unmatched_tags.discard('all')
                        unknown_tags = ((set(pb.only_tags) | set(pb.skip_tags)) -
                                        (matched_tags | unmatched_tags))

                        if unknown_tags:
                            continue

                    if options.listhosts:
                        print('  play #%d (%s): host count=%d' % (playnum, label, len(hosts)))
                        for host in hosts:
                            print('    %s' % host)

                    if options.listtasks:
                        print('  play #%d (%s):' % (playnum, label))

                        for task in play.tasks():
                            _only_tags = set(task.tags).intersection(pb.only_tags)
                            _skip_tags = set(task.tags).intersection(pb.skip_tags)
                            if (_only_tags and not _skip_tags):
                                if getattr(task, 'name', None) is not None:
                                    # meta tasks have no names
                                    print('    %s' % task.name)
                    print('')
                continue

            if options.syntax:
                # if we've not exited by now then we are fine.
                print('Playbook Syntax is fine')
                sys.exit(0)

            failed_hosts = []
            unreachable_hosts = []
            pb.run()

            hosts = sorted(pb.stats.processed.keys())
            callbacks.display(callbacks.banner("PLAY RECAP"))
            playbook_cb.on_stats(pb.stats)

            for h in hosts:
                t = pb.stats.summarize(h)
                if t['failures'] > 0:
                    failed_hosts.append(h)
                if t['unreachable'] > 0:
                    unreachable_hosts.append(h)

            retries = failed_hosts + unreachable_hosts

            if len(retries) > 0:
                filename = pb.generate_retry_inventory(retries)
                if filename:
                    callbacks.display("           to retry, use: --limit @%s\n" % filename)

            for h in hosts:
                t = pb.stats.summarize(h)

                callbacks.display("%s : %s %s %s %s" % (
                    hostcolor(h, t),
                    colorize('ok', t['ok'], 'green'),
                    colorize('changed', t['changed'], 'yellow'),
                    colorize('unreachable', t['unreachable'], 'red'),
                    colorize('failed', t['failures'], 'red')),
                    screen_only=True
                )

                callbacks.display("%s : %s %s %s %s" % (
                    hostcolor(h, t, False),
                    colorize('ok', t['ok'], None),
                    colorize('changed', t['changed'], None),
                    colorize('unreachable', t['unreachable'], None),
                    colorize('failed', t['failures'], None)),
                    log_only=True
                )

            print("")
            if len(failed_hosts) > 0:
                sys.exit(2)
            if len(unreachable_hosts) > 0:
                sys.exit(3)
    except errors.AnsibleError as e:
        callbacks.display("ERROR: %s" % e, color='red', stderr=True)
        sys.exit(1)


def run_cli(ctrl, name, argv):
    if name == 'ansible':
        return ansible_cmd(ctrl, argv)
    elif name == 'playbook':
        return playbook_cmd(ctrl, argv)
    raise NotImplementedError



from __future__ import unicode_literals
from ploy_ansible import ANSIBLE1, ANSIBLE2
import pytest
import sys


if sys.version_info < (3,):  # pragma: nocover
    too_view_arguments = 'too few arguments'
else:  # pragma: nocover
    too_view_arguments = 'the following arguments are required'


@pytest.fixture
def runmock(mock, monkeypatch):
    if ANSIBLE1:
        import ansible.playbook
        with mock.patch.object(ansible.playbook.PlayBook, "run", autospec=True) as runmock:
            yield runmock
    else:
        import ansible.executor.task_queue_manager
        with mock.patch.object(ansible.executor.task_queue_manager.TaskQueueManager, "run", autospec=True) as runmock:
            runmock.return_value = 0
            yield runmock


def test_register_ansible_module_path():
    pass


def test_register_ansible_module_path_from_multiple_entry_points():
    pass


def test_configure_without_args(capsys, ctrl, mock):
    with pytest.raises(SystemExit):
        ctrl(['./bin/ploy', 'configure'])
    (out, err) = capsys.readouterr()
    assert 'usage: ploy configure' in err
    assert too_view_arguments in err


def test_configure_with_nonexisting_instance(capsys, ctrl, mock):
    with pytest.raises(SystemExit):
        ctrl(['./bin/ploy', 'configure', 'bar'])
    (out, err) = capsys.readouterr()
    assert 'usage: ploy configure' in err
    assert "argument instance: invalid choice: 'bar'" in err


def test_configure_with_missing_yml(capsys, ctrl, mock):
    with pytest.raises(SystemExit):
        ctrl(['./bin/ploy', 'configure', 'foo'])
    (out, err) = capsys.readouterr()
    assert 'usage: ploy configure' in err
    assert "argument instance: invalid choice: 'foo'" in err


def test_configure_with_empty_yml(ctrl, mock, tempdir):
    tempdir['default-foo.yml'].fill('')
    with mock.patch('ploy_ansible.log') as LogMock:
        with pytest.raises(SystemExit):
            ctrl(['./bin/ploy', 'configure', 'foo'])
    assert len(LogMock.error.call_args_list) == 1
    call_args = LogMock.error.call_args_list[0][0]
    assert (
        "playbooks must be a list of plays" in call_args[0] or
        "playbooks must be formatted as a YAML list, got <type 'NoneType'>" in call_args[0] or
        "Empty playbook, nothing to do" in call_args[0])


def test_configure_asks_when_no_host_in_yml(ctrl, mock, tempdir):
    yml = tempdir['default-foo.yml']
    yml.fill([
        '---',
        '- {}'])
    with mock.patch('ploy_ansible.yesno') as YesNoMock:
        YesNoMock.return_value = False
        with pytest.raises(SystemExit):
            ctrl(['./bin/ploy', 'configure', 'foo'])
    assert len(YesNoMock.call_args_list) == 1
    call_args = YesNoMock.call_args_list[0][0]
    assert "Do you really want to apply '%s' to the host '%s'?" % (yml.path, 'default-foo') in call_args[0]


def test_configure(ctrl, runmock, tempdir):
    tempdir['default-foo.yml'].fill([
        '---',
        '- hosts: default-foo'])
    ctrl(['./bin/ploy', 'configure', 'foo'])
    assert runmock.called


def test_configure_playbook_option(ctrl, monkeypatch, ployconf, runmock, tempdir):
    from ploy_ansible import get_playbook
    yml = tempdir['default-bar.yml']
    yml.fill([
        '---',
        '- hosts: default-foo'])
    ployconf.fill([
        '[dummy-instance:foo]',
        'host = foo',
        'playbook = %s' % yml.path])
    playbooks = []

    def _get_playbook(self, *args, **kwargs):
        pb = get_playbook(self, *args, **kwargs)
        playbooks.append(pb)
        return pb

    monkeypatch.setattr("ploy_ansible.get_playbook", _get_playbook)
    ctrl(['./bin/ploy', 'configure', 'foo'])
    if ANSIBLE1:
        assert runmock.called
        assert runmock.call_args[0][0].filename == yml.path
    else:
        (pb,) = playbooks
        assert pb._file_name == yml.path


def test_configure_playbook_option_shadowing(ctrl, monkeypatch, ployconf, runmock, caplog, tempdir):
    from ploy_ansible import get_playbook
    yml_foo = tempdir['default-foo.yml']
    yml_foo.fill('')
    yml_bar = tempdir['default-bar.yml']
    yml_bar.fill([
        '---',
        '- hosts: default-foo'])
    ployconf.fill([
        '[dummy-instance:foo]',
        'host = foo',
        'playbook = %s' % yml_bar.path])
    playbooks = []

    def _get_playbook(self, *args, **kwargs):
        pb = get_playbook(self, *args, **kwargs)
        playbooks.append(pb)
        return pb

    monkeypatch.setattr("ploy_ansible.get_playbook", _get_playbook)
    ctrl(['./bin/ploy', 'configure', 'foo'])
    if ANSIBLE1:
        assert runmock.called
        assert runmock.call_args[0][0].filename == yml_bar.path
        assert [x.message for x in caplog.records] == [
            "Instance 'dummy-instance:foo' has the 'playbook' option set, but there is also a playbook at the default location '%s', which differs from '%s'." % (yml_foo.path, yml_bar.path),
            "Using playbook at '%s'." % yml_bar.path]
    else:
        (pb,) = playbooks
        assert pb._file_name == yml_bar.path
        assert [x.message for x in caplog.records] == [
            "Instance 'dummy-instance:foo' has the 'playbook' option set, but there is also a playbook at the default location '%s', which differs from '%s'." % (yml_foo.path, yml_bar.path),
            "Using playbook at '%s'." % yml_bar.path]


def test_configure_roles_option(ctrl, ployconf, runmock, tempdir):
    ployconf.fill([
        '[dummy-instance:foo]',
        'host = foo',
        'roles = ham egg'])
    tempdir['roles/egg/tasks/main.yml'].fill([])
    tempdir['roles/ham/tasks/main.yml'].fill([])
    ctrl(['./bin/ploy', 'configure', 'foo'])
    assert runmock.called
    if ANSIBLE1:
        assert runmock.call_args[0][0].filename in (
            "<dynamically generated from ['ham', 'egg']>",
            "<dynamically generated from [u'ham', u'egg']>")
        assert runmock.call_args[0][0].playbook == [{'hosts': ['default-foo'], 'user': 'root', 'roles': ['ham', 'egg']}]
        assert runmock.call_args[0][0].play_basedirs == [tempdir.directory]
    else:
        (call,) = runmock.call_args_list
        play = call[1]['play']
        assert play.hosts == ['default-foo']
        assert play.remote_user == 'root'
        assert [x.get_name() for x in play.get_roles()] == ['ham', 'egg']


def test_configure_roles_default_playbook_conflict(ctrl, ployconf, caplog, tempdir):
    yml = tempdir['default-foo.yml']
    yml.fill('')
    ployconf.fill([
        '[dummy-instance:foo]',
        'host = foo',
        'roles = ham egg'])
    with pytest.raises(SystemExit):
        ctrl(['./bin/ploy', 'configure', 'foo'])
    assert [x.message for x in caplog.records] == [
        "Using playbook at '%s'." % yml.path,
        "You can't use a playbook and the 'roles' options at the same time for instance 'dummy-instance:foo'."]


def test_configure_roles_playbook_option_conflict(ctrl, ployconf, caplog, tempdir):
    yml = tempdir['default-bar.yml']
    yml.fill([
        '---',
        '- hosts: default-foo'])
    ployconf.fill([
        '[dummy-instance:foo]',
        'playbook = %s' % yml.path,
        'host = foo',
        'roles = ham egg'])
    with pytest.raises(SystemExit):
        ctrl(['./bin/ploy', 'configure', 'foo'])
    assert [x.message for x in caplog.records] == [
        "Using playbook at '%s'." % yml.path,
        "You can't use a playbook and the 'roles' options at the same time for instance 'dummy-instance:foo'."]


def test_configure_playbook_discovery(ctrl, ployconf, caplog, runmock, tempdir):
    tempdir['foo.yml'].fill('')
    yml = tempdir['default-foo.yml']
    yml.fill([
        '---',
        '- hosts: default-foo'])
    ctrl(['./bin/ploy', 'configure', 'foo'])
    assert runmock.called
    assert [x.message for x in caplog.records] == [
        "Using playbook at '%s'." % yml.path]


def test_configure_with_extra_vars(ctrl, runmock, tempdir):
    tempdir['default-foo.yml'].fill([
        '---',
        '- hosts: default-foo'])
    ctrl(['./bin/ploy', 'configure', 'foo', '-e', 'foo=bar', '-e', 'bar=foo'])
    assert runmock.called
    if ANSIBLE1:
        assert runmock.call_args[0][0].extra_vars == dict(foo='bar', bar='foo')
    else:
        (call,) = runmock.call_args_list
        play = call[1]['play']
        variables = play.get_variable_manager().get_vars()
        assert variables['foo'] == 'bar'
        assert variables['bar'] == 'foo'


def test_configure_with_tags(ctrl, runmock, tempdir):
    tempdir['default-foo.yml'].fill([
        '---',
        '- hosts: default-foo'])
    ctrl(['./bin/ploy', 'configure', 'foo', '-t', 'ham,egg'])
    assert runmock.called
    if ANSIBLE1:
        play = runmock.call_args[0][0]
        assert sorted(play.only_tags) == ['egg', 'ham']
        assert play.skip_tags == []
    else:
        (call,) = runmock.call_args_list
        tqm = call[0][0]
        assert sorted(tqm._options.tags) == ['egg', 'ham']
        assert tqm._options.skip_tags == []


def test_configure_with_skip_tags(ctrl, runmock, tempdir):
    tempdir['default-foo.yml'].fill([
        '---',
        '- hosts: default-foo'])
    ctrl(['./bin/ploy', 'configure', 'foo', '--skip-tags', 'ham,egg'])
    assert runmock.called
    if ANSIBLE1:
        play = runmock.call_args[0][0]
        assert play.only_tags == ['all']
        assert sorted(play.skip_tags) == ['egg', 'ham']
    else:
        (call,) = runmock.call_args_list
        tqm = call[0][0]
        assert tqm._options.tags == ['all']
        assert sorted(tqm._options.skip_tags) == ['egg', 'ham']


def test_playbook_without_args(capsys, ctrl, mock):
    with pytest.raises(SystemExit):
        ctrl(['./bin/ploy', 'playbook'])
    (out, err) = capsys.readouterr()
    if ANSIBLE1:
        assert 'Usage: ploy playbook playbook.yml' in err
    else:
        assert 'Usage:' in out
        assert '[options] playbook.yml' in out


def test_playbook_with_nonexisting_playbook(capsys, ctrl, mock):
    with pytest.raises(SystemExit):
        ctrl(['./bin/ploy', 'playbook', 'bar.yml'])
    (out, err) = capsys.readouterr()
    assert "the playbook: bar.yml could not be found" in err


def test_playbook_with_empty_yml(capsys, ctrl, mock, tempdir):
    yml = tempdir['foo.yml']
    yml.fill('')
    with pytest.raises(SystemExit):
        ctrl(['./bin/ploy', 'playbook', yml.path])
    (out, err) = capsys.readouterr()
    assert (
        "playbooks must be a list of plays" in err or
        "playbooks must be formatted as a YAML list, got <type 'NoneType'>" in err)


def test_playbook_asks_when_no_host_in_yml(capsys, ctrl, mock, tempdir):
    yml = tempdir['foo.yml']
    yml.fill([
        '---',
        '- {}'])
    with pytest.raises(SystemExit):
        ctrl(['./bin/ploy', 'playbook', yml.path])
    (out, err) = capsys.readouterr()
    assert (
        "'hosts' is required but was not set" in err or
        'hosts declaration is required' in err)


def test_playbook(ctrl, mock, monkeypatch, runmock, tempdir):
    yml = tempdir['foo.yml']
    yml.fill([
        '---',
        '- hosts: foo'])
    if ANSIBLE2:
        runmock = mock.MagicMock()
        monkeypatch.setattr("ansible.cli.playbook.PlaybookCLI.run", runmock)
        runmock.return_value = 0
    ctrl(['./bin/ploy', 'playbook', yml.path])
    assert runmock.called


def test_ansible_without_args(capsys, ctrl, mock):
    with pytest.raises(SystemExit):
        ctrl(['./bin/ploy', 'ansible'])
    (out, err) = capsys.readouterr()
    assert 'Usage: ' in out
    assert '<host-pattern> [options]' in out


def test_ansible_with_nonexisting_instance(capsys, ctrl, mock):
    with pytest.raises(SystemExit):
        ctrl(['./bin/ploy', 'ansible', 'bar'])
    (out, err) = capsys.readouterr()
    assert "No hosts matched" in err


def test_ansible(ctrl, mock, monkeypatch):
    runmock = mock.MagicMock()
    if ANSIBLE1:
        monkeypatch.setattr("ansible.runner.Runner.run", runmock)
        runmock.return_value = dict(
            contacted=dict(),
            dark=[])
    else:
        monkeypatch.setattr("ansible.cli.adhoc.AdHocCLI.run", runmock)
        runmock.return_value = 0
    ctrl(['./bin/ploy', 'ansible', 'default-foo', '-a', 'ls'])
    assert runmock.called


class PasswordDeleteError(Exception):
    pass


class MockKeyringErrors:
    PasswordDeleteError = PasswordDeleteError


class MockKeyring:
    errors = MockKeyringErrors()

    def __init__(self):
        self.passwords = {}

    def delete_password(self, system, username):
        if system == "ploy_ansible":
            if username in self.passwords:
                del self.passwords[username]
                return
        raise PasswordDeleteError()

    def get_password(self, system, username):
        if system == "ploy_ansible":
            return self.passwords.get(username)

    def set_password(self, system, username, password):
        if system == "ploy_ansible":
            self.passwords[username] = password


@pytest.fixture
def getpass(mock, monkeypatch):
    getpass = mock.MagicMock()
    monkeypatch.setattr("getpass.getpass", getpass)
    return getpass


@pytest.fixture
def keyring(monkeypatch):
    keyring = MockKeyring()
    monkeypatch.setattr("ploy_ansible.KeyringSource.keyring", keyring)
    return keyring


def test_vault_key_generate(ctrl, keyring, ployconf):
    ployconf.fill([
        '[ansible]',
        'vault-password-source = ploy_ansible-test-key'])
    assert 'ploy_ansible-test-key' not in keyring.passwords
    ctrl(['./bin/ploy', 'vault-key', 'generate'])
    assert 'ploy_ansible-test-key' in keyring.passwords


def test_vault_key_set(ctrl, getpass, keyring, ployconf):
    getpass.return_value = "foo"
    ployconf.fill([
        '[ansible]',
        'vault-password-source = ploy_ansible-test-key'])
    assert 'ploy_ansible-test-key' not in keyring.passwords
    ctrl(['./bin/ploy', 'vault-key', 'set'])
    assert keyring.passwords['ploy_ansible-test-key'] == "foo"


def test_vault_key_set_existing(ctrl, getpass, keyring, mock, ployconf):
    getpass.return_value = "foo"
    ployconf.fill([
        '[ansible]',
        'vault-password-source = ploy_ansible-test-key'])
    keyring.passwords['ploy_ansible-test-key'] = "bar"
    with mock.patch('ploy_ansible.yesno') as YesNoMock:
        YesNoMock.return_value = False
        ctrl(['./bin/ploy', 'vault-key', 'set'])
    assert keyring.passwords['ploy_ansible-test-key'] == "bar"
    with mock.patch('ploy_ansible.yesno') as YesNoMock:
        YesNoMock.return_value = True
        ctrl(['./bin/ploy', 'vault-key', 'set'])
    assert keyring.passwords['ploy_ansible-test-key'] == "foo"


def test_vault_key_delete(ctrl, keyring, mock, ployconf):
    ployconf.fill([
        '[ansible]',
        'vault-password-source = ploy_ansible-test-key'])
    keyring.passwords['ploy_ansible-test-key'] = "foo"
    with mock.patch('ploy_ansible.yesno') as YesNoMock:
        YesNoMock.return_value = False
        ctrl(['./bin/ploy', 'vault-key', 'delete'])
    assert 'ploy_ansible-test-key' in keyring.passwords
    with mock.patch('ploy_ansible.yesno') as YesNoMock:
        YesNoMock.return_value = True
        ctrl(['./bin/ploy', 'vault-key', 'delete'])
    assert 'ploy_ansible-test-key' not in keyring.passwords



from __future__ import print_function
from ansible import errors
from ansible.plugins.connection import ConnectionBase
from execnet.gateway_base import _Serializer
import execnet
import os
import paramiko
import ploy_ansible
import ploy_ansible.remote
import sys

try:
    from __main__ import display
except ImportError:
    from ansible.utils.display import Display
    display = Display()


# monkey patch execnet serializer to support AnsibleUnsafeText
if not hasattr(_Serializer, 'save_AnsibleUnsafeText'):
    def save_AnsibleUnsafeText(self, s):
        self._save(u"%s" % s)
    _Serializer.save_AnsibleUnsafeText = save_AnsibleUnsafeText


class RPCWrapper(object):

    def __init__(self, channel):
        self.channel = channel

    def __getattr__(self, name):
        def call(*args, **kw):
            self.channel.send((name, args, kw))
            result = self.channel.receive()
            try:
                result[0]
            except (TypeError, IndexError):
                pass
            else:
                if result[0] == 'remote-core-error':
                    print(result[1], file=sys.stderr)
                    raise RuntimeError('Remote exception encountered.')
            return result
        return call


class SSHArgs:
    def __init__(self, args):
        self.args = args

    def split(self):
        return self.args


class Connection(ConnectionBase):
    ''' execnet based connections '''

    transport = 'execnet'

    def __init__(self, *args, **kwargs):
        super(Connection, self).__init__(*args, **kwargs)

        self.host = self._play_context.remote_addr
        if hasattr(self.host, 'instance'):
            self.host = self.host.instance.uid
        self.port = self._play_context.port
        self.user = self._play_context.remote_user
        self.has_pipelining = False
        self.become_methods_supported = ['sudo']
        self._cache_key = (self.host, self.port, self.user)

    def _connect(self):
        if self._cache_key not in ploy_ansible.RPC_CACHE:
            ctrl = self._play_context._ploy_ctrl
            instance = ctrl.instances[self.host]
            if hasattr(instance, '_status'):
                if instance._status() != 'running':
                    raise errors.AnsibleError("Instance '%s' unavailable." % instance.config_id)
            try:
                ssh_info = instance.init_ssh_key(user=self.user)
            except paramiko.SSHException as e:
                raise errors.AnsibleError("Couldn't validate fingerprint for '%s': %s" % (instance.config_id, e))
            spec = execnet.XSpec('ssh')
            ssh_args = instance.ssh_args_from_info(ssh_info)
            if display.verbosity > 3:
                ssh_args += ["-vvv"]
            spec.ssh = SSHArgs(ssh_args)
            vars = instance.get_ansible_variables()
            spec.python = vars.get('ansible_python_interpreter', 'python')
            gw = execnet.makegateway(spec)
            try:
                channel = gw.remote_exec(ploy_ansible.remote)
            except IOError as e:
                raise errors.AnsibleError("Couldn't open execnet channel for '%s': %s" % (instance.config_id, e))
            ploy_ansible.RPC_CACHE[self._cache_key] = RPCWrapper(channel)
        self.rpc = ploy_ansible.RPC_CACHE[self._cache_key]
        return self

    def exec_command(self, cmd, in_data=None, sudoable=True):
        super(Connection, self).exec_command(cmd, in_data=in_data, sudoable=sudoable)

        if in_data is not None:
            raise errors.AnsibleError("Internal Error: this module does not support optimized module pipelining")

        display.vvv("execnet exec_command %s" % cmd)
        rc, stdout, stderr = self.rpc.exec_command(cmd)
        return (rc, stdout, stderr)

    def put_file(self, in_path, out_path):
        display.vvv("execnet put_file %r %r" % (in_path, out_path))
        if not os.path.exists(in_path):
            raise errors.AnsibleFileNotFound("file or module does not exist: %s" % in_path)
        with open(in_path) as f:
            self.rpc.put_file(f.read(), out_path)

    def fetch_file(self, in_path, out_path):
        display.vvv("execnet fetch_file %r %r" % (in_path, out_path))
        data = self.rpc.fetch_file(in_path)
        with os.path.open(out_path, 'w') as f:
            f.write(data)

    def close(self):
        pass



import setuptools
import os
import sys


here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
HISTORY = open(os.path.join(here, 'HISTORY.rst')).read()


version = "2.0.0b6"


classifiers = [
    'Environment :: Console',
    'Intended Audience :: System Administrators',
    'Programming Language :: Python :: 2.7',
    'Topic :: System :: Installation/Setup',
    'Topic :: System :: Systems Administration']

extras_require = {}

install_requires = [
    'execnet',
    'ploy >= 2.0.0b3']


def get_environment_marker_support_level():
    """
    Tests how well setuptools supports PEP-426 environment marker.

    The first known release to support it is 0.7 (and the earliest on PyPI seems to be 0.7.2
    so we're using that), see: https://setuptools.readthedocs.io/en/latest/history.html#id350

    The support is later enhanced to allow direct conditional inclusions inside install_requires,
    which is now recommended by setuptools. It first appeared in 36.2.0, went broken with 36.2.1, and
    again worked since 36.2.2, so we're using that. See:
    https://setuptools.readthedocs.io/en/latest/history.html#v36-2-2
    https://github.com/pypa/setuptools/issues/1099

    References:

    * https://wheel.readthedocs.io/en/latest/index.html#defining-conditional-dependencies
    * https://www.python.org/dev/peps/pep-0426/#environment-markers
    * https://setuptools.readthedocs.io/en/latest/setuptools.html#declaring-platform-specific-dependencies
    """
    import pkg_resources
    try:
        version = pkg_resources.parse_version(setuptools.__version__)
        if version >= pkg_resources.parse_version('36.2.2'):
            return 2
        if version >= pkg_resources.parse_version('0.7.2'):
            return 1
    except Exception as exc:
        sys.stderr.write("Could not test setuptool's version: %s\n" % exc)
    return 0


if get_environment_marker_support_level() >= 2:
    install_requires.append('ansible>=1.9,!=2.0.*,!=2.1.*,!=2.2.*,!=2.3.*;python_version<"3.0"')
    install_requires.append('ansible>=2.4;python_version>"3.0"')
    classifiers.extend([
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'])
elif get_environment_marker_support_level() == 1:
    extras_require[':python_version<"3.0"'] = ['ansible>=1.9,!=2.0.*,!=2.1.*,!=2.2.*,!=2.3.*']
    extras_require[':python_version>"3.0"'] = ['ansible>=2.4']
    classifiers.extend([
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'])
else:
    if sys.version_info < (3, 0):
        install_requires.append('ansible>=1.9,!=2.0.*,!=2.1.*,!=2.2.*,!=2.3.*')
    else:
        install_requires.append('ansible>=2.4')


setuptools.setup(
    version=version,
    description="Plugin to integrate Ansible with ploy.",
    long_description=README + "\n\n" + HISTORY,
    name="ploy_ansible",
    author='Florian Schulze',
    author_email='florian.schulze@gmx.net',
    license="GPLv3",  # infected by code for ansible and playbook commands
    url='http://github.com/ployground/ploy_ansible',
    classifiers=classifiers,
    include_package_data=True,
    zip_safe=False,
    packages=['ploy_ansible'],
    install_requires=install_requires,
    extras_require=extras_require,
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*',
    entry_points="""
        [ansible_paths]
        ploy_ansible = ploy_ansible:ansible_paths
        [ploy.plugins]
        ansible = ploy_ansible:plugin
    """)



from __future__ import print_function
from ansible import errors
from ansible import utils
from ansible.callbacks import vvv
import execnet
import os
import paramiko
import pipes
import ploy_ansible
import ploy_ansible.remote
import sys


class RPCWrapper(object):

    def __init__(self, channel):
        self.channel = channel

    def __getattr__(self, name):
        def call(*args, **kw):
            self.channel.send((name, args, kw))
            result = self.channel.receive()
            try:
                result[0]
            except (TypeError, IndexError):
                pass
            else:
                if result[0] == 'remote-core-error':
                    print(result[1], file=sys.stderr)
                    raise RuntimeError('Remote exception encountered.')
            return result
        return call


class SSHArgs:
    def __init__(self, args):
        self.args = args

    def split(self):
        return self.args


class Connection(object):
    ''' execnet based connections '''

    def __init__(self, runner, host, port, user, password, private_key_file, *args, **kwargs):
        self.runner = runner
        self.host = host
        self.user = user
        self.has_pipelining = False
        self.become_methods_supported = ['sudo']
        self._cache_key = (host, user)

    def connect(self):
        if self._cache_key not in ploy_ansible.RPC_CACHE:
            ctrl = self.runner._ploy_ctrl
            instance = ctrl.instances[self.host]
            if hasattr(instance, '_status'):
                if instance._status() != 'running':
                    raise errors.AnsibleError("Instance '%s' unavailable." % instance.config_id)
            try:
                ssh_info = instance.init_ssh_key(user=self.user)
            except paramiko.SSHException as e:
                raise errors.AnsibleError("Couldn't validate fingerprint for '%s': %s" % (instance.config_id, e))
            spec = execnet.XSpec('ssh')
            ssh_args = instance.ssh_args_from_info(ssh_info)
            if utils.VERBOSITY > 3:
                ssh_args += ["-vvv"]
            spec.ssh = SSHArgs(ssh_args)
            vars = instance.get_ansible_variables()
            spec.python = vars.get('ansible_python_interpreter', 'python')
            gw = execnet.makegateway(spec)
            try:
                channel = gw.remote_exec(ploy_ansible.remote)
            except IOError as e:
                raise errors.AnsibleError("Couldn't open execnet channel for '%s': %s" % (instance.config_id, e))
            ploy_ansible.RPC_CACHE[self._cache_key] = RPCWrapper(channel)
        self.rpc = ploy_ansible.RPC_CACHE[self._cache_key]
        return self

    def exec_command(self, cmd, tmp_path, become_user, **kwargs):
        executable = kwargs.get('executable', '/bin/sh')
        sudoable = kwargs.get('sudoable', False)
        if 'su' in kwargs or 'su_user' in kwargs:
            raise errors.AnsibleError("Internal Error: this module does not support running commands via su")

        if kwargs.get('in_data'):
            raise errors.AnsibleError("Internal Error: this module does not support optimized module pipelining")

        if sudoable and getattr(self.runner, 'become', False) and self.runner.become_method not in self.become_methods_supported:
            raise errors.AnsibleError("Internal Error: this module does not support running commands via %s" % self.runner.become_method)

        become = getattr(self.runner, 'become', False) or getattr(self.runner, 'sudo', False)
        remote_cmd = []
        if not become or not sudoable:
            if executable:
                remote_cmd.append(executable + ' -c ' + pipes.quote(cmd))
            else:
                remote_cmd.append(cmd)
        else:
            if hasattr(self.runner, 'become_exe'):
                becomecmd, prompt, success_key = utils.make_become_cmd(cmd, become_user, executable, self.runner.become_method, '', self.runner.become_exe)
            elif hasattr(self.runner, 'sudo_exe'):
                becomecmd, prompt, success_key = utils.make_sudo_cmd(self.runner.sudo_exe, become_user, executable, cmd)
            else:
                becomecmd, prompt, success_key = utils.make_sudo_cmd(become_user, executable, cmd)
            remote_cmd.append(becomecmd)
        remote_cmd = ' '.join(remote_cmd)
        vvv("execnet exec_command %r" % remote_cmd)
        rc, stdout, stderr = self.rpc.exec_command(remote_cmd)
        return (rc, '', stdout, stderr)

    def put_file(self, in_path, out_path):
        vvv("execnet put_file %r %r" % (in_path, out_path))
        if not os.path.exists(in_path):
            raise errors.AnsibleFileNotFound("file or module does not exist: %s" % in_path)
        with open(in_path) as f:
            self.rpc.put_file(f.read(), out_path)

    def fetch_file(self, in_path, out_path):
        vvv("execnet fetch_file %r %r" % (in_path, out_path))
        data = self.rpc.fetch_file(in_path)
        with os.path.open(out_path, 'w') as f:
            f.write(data)

    def close(self):
        pass
